import numpy as np
from sklearn.metrics import pairwise_distances


def energy_distance(x, y):
    x = np.reshape(x, (len(x), -1))
    y = np.reshape(y, (len(y), -1))
    n = x.shape[0]
    a = np.mean(pairwise_distances(x, y, metric='euclidean'))
    b = np.mean(pairwise_distances(y, metric='euclidean'))
    c = np.mean(pairwise_distances(x, metric='euclidean'))
    return 2 * a - b - c



from .mmd import maximum_mean_discrepancy
from .fd import frechet_distance
from .energy import energy_distance

__all__ = [
    'maximum_mean_discrepancy',
    'frechet_distance',
    'energy_distance'
]






from ..algorithms.cpdc import ChangePointDetectionBase
from roerich.scores.fd import frechet_distance
from roerich.scores.mmd import maximum_mean_discrepancy
from roerich.scores.energy import energy_distance
import numpy as np
from sklearn.utils import resample


class SlidingWindows(ChangePointDetectionBase):

    def __init__(self, metric='mmd', bootstrap=False, periods=1, window_size=100, step=1, n_runs=1):
        super().__init__(periods=periods, window_size=window_size, step=step, n_runs=n_runs)
        self.metric = metric
        self.bootstrap = bootstrap

        """
        Sliding windows change point detection method

        Parameters:
        -----------
        
        metric: str/function, default=None
            Possible values: {'energy', 'fd', 'mmd'}.
        'energy' - energy distance
        'fd' - Frechet distance
        'mmd' - maximum mean discrepancy
        Otherwise, a function of format like in frechet_distance from roerich.scores.fd
        
        bootstrap: bool, default=False
            Whether to use bootstrapping procedure. Resampling is done n_runs times and the average score is returned
        
        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for classification. In the most cases periods=1
        will be a good choice.

        window_size: int, default=100
            Number of consecutive observations of a time series in test and reference
        windows. Recommendation: select the value so that there is only one change point within 2*window_size
        observations of the signal.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_runs: int, default=1
            Number of times the bootstrapping is applied
            
        """

    def reference_test_predict(self, X_ref, X_test):

        if self.metric == "energy":
            return energy_distance(X_ref, X_test)
        elif self.metric == "fd":
            return frechet_distance(X_ref, X_test)
        elif self.metric == "mmd":
            return maximum_mean_discrepancy(X_ref, X_test)
        elif callable(self.metric):
            return self.metric(X_ref, X_test)
        else:
            raise ValueError("metric should be one of: energy, fd, mmd; or a function should be "
                             "passed")

    def reference_test_predict_n_times(self, X_ref, X_test):
        """
        Estimate change point detection score several times for a pair of test and reference windows
        where bootstrapping procedure is applied to the test window

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Returns:
        --------
        score: float
            Estimated average change point detection score for a pair of window.
        """

        scores = []
        for i in range(self.n_runs):
            if self.bootstrap:
                X_ref_b, X_test_b = resample(X_ref), resample(X_test)
            else:
                X_ref_b, X_test_b = X_ref, X_test
            ascore = self.reference_test_predict(X_ref_b, X_test_b)
            scores.append(ascore)

        return np.mean(scores)



import numpy as np



def generate_dataset(period=200, N_tot=1000):
    mu = 0
    sigma = 1.
    N = 1

    T = [0, 1]
    X = [np.random.normal(mu, sigma, 1)[0], np.random.normal(mu, sigma, 1)[0]]
    cps = []

    for i in range(2, N_tot):
        if i % period == 0:
            N += 1
            mu += 0.5 * N
            cps.append(i)
        T += [i]
        ax = 0.6 * X[i-1] - 0.5 * X[i-2] + np.random.normal(mu, sigma, 1)[0]
        X += [ax]
    return np.array(X).reshape(-1, 1), cps



import numpy as np
import pytest
from roerich import explanation
from roerich import change_point

def test_matrix_importance():
    # generate time series
    n = 400
    X1 = np.random.normal([0, 0, 0], 1, size=(n, 3))
    X2 = np.random.normal([0, 1, 2], 1, size=(n, 3))
    X = np.concatenate((X1, X2), axis=0)
    cps_true = [n]

    # base detection algo
    cp = change_point.EnergyDistanceCalculator()

    # importance
    mm = explanation.MatrixImportance(cp)
    score, cps_pred = mm.predict(X)
    mat_u = mm.predict_union(X)
    mat_e = mm.predict_exclude(X)

    assert score.shape == (2 * n,)
    assert mat_u.shape == (2*n, 3, 3)
    assert mat_e.shape == (2 * n, 3, 3)



import numpy as np
from sklearn import metrics


def maximum_mean_discrepancy(X, Y):
    '''
    [Computationally expensive. Recommended for samples size < 5000]
    Calculates the Maximum Mean Discrepancy between X and Y samples.

    Parameters:
    -----------
    X: numpy.ndarray of shape [n_samples, n_features]
        The first sample.
    Y: numpy.ndarray of shape [n_samples, n_features]
        The second sample.

    Return:
    -------
    distance: float
        The estimated Frechet Distance.
    '''

    # 1d array to nd one
    X = np.reshape(X, (len(X), -1))
    Y = np.reshape(Y, (len(Y), -1))

    agg_matrix = np.concatenate((X, Y), axis=0)
    distances = metrics.pairwise_distances(agg_matrix)
    median_distance = np.median(distances) + 10**-10
    gamma = 1.0 / (2 * median_distance**2)

    XX = metrics.pairwise.rbf_kernel(X, X, gamma)
    YY = metrics.pairwise.rbf_kernel(Y, Y, gamma)
    XY = metrics.pairwise.rbf_kernel(X, Y, gamma)

    mmd = XX.mean() + YY.mean() - 2 * XY.mean()

    return mmd



import numpy as np
import pytest
from inspect import getmembers, isfunction, isclass
from roerich import change_point
from sklearn.linear_model import LogisticRegression
from roerich.scores import frechet_distance

models = [f[1] for f in getmembers(change_point, isclass)]

@pytest.mark.parametrize("model", models)
def test_base(model):
    # generate time series
    n = 400
    X1 = np.random.normal(0, 1, size=(n, 1))
    X2 = np.random.normal(2, 1, size=(n, 1))
    X = np.concatenate((X1, X2), axis=0)
    cps_true = [n]

    # detection
    cp = model()
    score, cps_pred = cp.predict(X)

cpdc = [change_point.ChangePointDetectionClassifier,
        change_point.ChangePointDetectionClassifierCV]
@pytest.mark.parametrize("model", cpdc)
def test_base_classifier(model):
    # generate time series
    n = 200
    X1 = np.random.normal(0, 1, size=(n, 1))
    X2 = np.random.normal(2, 1, size=(n, 1))
    X = np.concatenate((X1, X2), axis=0)
    cps_true = [n]

    # test clf
    lr = LogisticRegression()
    for clf in ['logreg', 'qda', 'dt', 'rf', 'mlp', 'knn', 'nb', lr]:
        cp = model(base_classifier=clf, window_size=50)
        score, cps_pred = cp.predict(X)

@pytest.mark.parametrize("model", cpdc)
def test_metric(model):
    # generate time series
    n = 200
    X1 = np.random.normal(0, 1, size=(n, 1))
    X2 = np.random.normal(2, 1, size=(n, 1))
    X = np.concatenate((X1, X2), axis=0)
    cps_true = [n]

    # test clf
    fd = frechet_distance
    for m in ['klsym', 'pesym', 'jsd', 'mmd', 'fd', fd]:
        cp = model(base_classifier='logreg', window_size=50, metric=m)
        score, cps_pred = cp.predict(X)



import numpy as np
import pytest
from scipy.stats import norm, uniform
from inspect import getmembers, isfunction, isclass
from roerich import density_ratio

models = [f[1] for f in getmembers(density_ratio, isclass)]

@pytest.mark.parametrize("model", models)
def test_base(model):
    # generate samples
    N = 1000
    p0 = uniform(-5, 10)
    p1 = norm(0, 1)
    X = np.concatenate((p0.rvs((N, 1)), p1.rvs((N, 1))))
    y = np.array([0] * N + [1] * N)

    # true density ratio
    true_ratio = p1.pdf(X) / p0.pdf(X)

    # direct density ratio estimation
    reg = model()
    reg.fit(X, y)
    pred_ratio = reg.predict_proba_ratio(X)



from abc import ABCMeta, abstractmethod
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from copy import deepcopy
from joblib import Parallel, delayed
from scipy import interpolate
from scipy.signal import argrelmax, savgol_filter

from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB

from sklearn.utils._testing import ignore_warnings
from sklearn.exceptions import ConvergenceWarning

from roerich.scores import maximum_mean_discrepancy, frechet_distance

from roerich.algorithms.models import GBDTRuLSIFRegressor, NNRuLSIFRegressor


# utils
def autoregression_matrix(X, periods=1, fill_value=0):
    shifted_x = [pd.DataFrame(X).shift(periods=i, fill_value=fill_value).values for i in range(periods)]
    X_auto = np.hstack(tuple(shifted_x))
    return X_auto


def reference_test(X, window_size=2, step=1):
    T = []
    reference = []
    test = []
    for i in range(2 * window_size - 1, len(X), step):
        T.append(i)
        reference.append(X[i - 2 * window_size + 1:i - window_size + 1])
        test.append(X[i - window_size + 1:i + 1])
    return np.array(T), np.array(reference), np.array(test)


def KL(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10 ** -3)) - np.mean(np.log(1. - test_preds + 10 ** -3))


def KL_sym(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10 ** -3)) - np.mean(np.log(1. - test_preds + 10 ** -3)) + \
        np.mean(np.log(1. - ref_preds + 10 ** -3)) - np.mean(np.log(ref_preds + 10 ** -3))


def JSD(ref_preds, test_preds):
    return np.log(2) + 0.5 * np.mean(np.log(test_preds + 10 ** -3)) + 0.5 * np.mean(np.log(1. - ref_preds + 10 ** -3))


def PE(ref_preds, test_preds):
    scores = test_preds / (1. - test_preds + 10 ** -6) - 1.
    scores = np.clip(scores, 0, 1000)
    return np.mean(scores)


def PE_sym(ref_preds, test_preds):
    scores_1 = test_preds / (1. - test_preds + 10 ** -6) - 1.
    scores_1 = np.clip(scores_1, 0, 1000)
    scores_2 = (1. - ref_preds) / (ref_preds + 10 ** -6) - 1.
    scores_2 = np.clip(scores_2, 0, 1000)
    return np.mean(scores_1) + np.mean(scores_2)


# Base
class ChangePointDetectionBase(metaclass=ABCMeta):

    def __init__(self, periods=1, window_size=100, step=1, n_runs=1):
        """
        Change point detection algorithm based on binary classification.

        Parameters:
        -----------
        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for classification. In the most cases periods=1
        will be a good choice.

        window_size: int, default=100
            Number of consecutive observations of a time series in test and reference
        windows. Recommendation: select the value so that there is only one change point within 2*window_size
        observations of the signal.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_runs: int, default=1
            Number of times, the binary classifier runs on each pair of test and reference
        windows. Observations in the windows are divided randomly between train and validation sample every time.
        n_runs > 1 helps to reduce noise in the change point detection score.
        
        """

        self.periods = periods
        self.window_size = window_size
        self.step = step
        self.n_runs = n_runs

    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Returns:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        score = 0

        return score

    def reference_test_predict_n_times(self, X_ref, X_test):
        """
        Estimate change point detection score several times for  a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Returns:
        --------
        score: float
            Estimated average change point detection score for a pair of window.
        """

        scores = []
        for i in range(self.n_runs):
            ascore = self.reference_test_predict(X_ref, X_test)
            scores.append(ascore)

        return np.mean(scores)

    def postprocessing(self, T, T_score, score):
        """
        Interpolates and shifts a change point detection score, estimates peak positions.
        
        Parameters:
        -----------
        T: numpy.array
            A broader time-step interval
        T_score: numpy.array
            A time intervals of CPD scores
        score: numpy.array
            A CPD scores

        Returns:
        --------
        new_score: numpy.array
            Interpolated and shifted CPD scores.
        peaks: numpy.array
            Positions of peaks in the CPD scores.
        """

        inter = interpolate.interp1d(T_score - self.window_size, score,
                                     kind='linear', fill_value=(0, 0), bounds_error=False)
        score_new = inter(T)
        length = max(min(21, self.window_size//2), 5)
        score_smooth = savgol_filter(score_new, length, 3)
        peaks = argrelmax(score_smooth, order=self.window_size)[0]

        return score_smooth, peaks

    def predict(self, X):
        """
        Estimate change point detection score for a time series.

        Parameters:
        -----------
        X: numpy.ndarray
            Time series observation.

        Retunrs:
        --------
        T_score: numpy.array
            Array of timestamps.
        scores: numpy.array
            Estimated change point detection score.
        """

        # data preprocessing
        X_auto = autoregression_matrix(X, periods=self.periods, fill_value=0)
        T, reference, test = reference_test(X_auto, window_size=self.window_size, step=1)

        # change point score estimation
        iters = range(0, len(reference), self.step)
        score = Parallel(n_jobs=-1)(delayed(self.reference_test_predict_n_times)(reference[i], test[i]) for i in iters)
        T_score = np.array([T[i] for i in iters])

        # cpd score postprocessing
        new_score, peaks = self.postprocessing(np.arange(len(X)), T_score, score)

        return new_score, peaks


# Classification
class ChangePointDetectionClassifier(ChangePointDetectionBase):

    def __init__(self, base_classifier='mlp', metric="klsym",
                 periods=1, window_size=100, step=1, n_runs=1):
        """

        Change point detection algorithm based on binary classification [1]. It takes two sliding windows
        (reference and test) in a signal, and separate them using the classifier. The classification quality is
        considered as a change point detection score.

        [1] Mikhail Hushchyn and Andrey Ustyuzhanin. “Generalization of Change-Point Detection in Time Series Data Based on Direct Density Ratio Estimation.” J. Comput. Sci. 53 (2021): 101385.

        Parameters:
        -----------
        base_classifier: {'logreg', 'qda', 'dt', 'rf', 'mlp', 'knn', 'nb'} or callable, default='mlp'
            Sklearn-like binary classifier to separate reference and test sliding windows in the signal.

            - 'logreg', Logistic Regression classifier,
              sklearn.linear_model.LogisticRegression()

            - 'qda', Quadratic Discriminant Analysis classifier,
              sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis(store_covariance=True, reg_param=0.01)

            - 'dt', Decision Tree classifier,
              sklearn.tree.DecisionTreeClassifier(min_samples_leaf=10, max_depth=6)

            - 'rf', Random Forest classifier,
              sklearn.ensemble.RandomForestClassifier((n_estimators=100, min_samples_leaf=10))

            - 'mlp', Multilayer Perceptron classifier,
              sklearn.neural_network.MLPClassifier(hidden_layer_sizes=(100,100), solver="adam", activation="relu", learning_rate_init=0.1, max_iter=50, alpha=1.)

            - 'knn', K Neighbors classifier,
              sklearn.neighbors.KNeighborsClassifier(n_neighbors=10)

            - 'nb', Gaussian Naive Bayes classifier,
              sklearn.naive_bayes.GaussianNB(var_smoothing=0.01)

            - Callable sklearn-like classifier,
              Example: base_classifier = LogisticRegression()

        metric: {'klsym', 'pesym', 'jsd', 'mmd', 'fd'} or callable, default='klsym'.
            {'KL', 'PE'} will be deprecated in future versions.
            Name of a score function, that is used to measure the classifier quality based on predictions
            for reference (p_ref) and test (p_test) windows. It is considered as change point detection score.

            - 'klsym' or 'KL_sym', symmetric Kullback-Leibler (KL) divergence,
            KL(p_test||p_ref) + KL(p_ref||p_test)

            - 'pesym' or 'PE_sym', symmetric Pearson (PE) divergence,
            PE(p_test||p_ref) + PE(p_ref||p_test)

            - 'jsd' or 'JSD', Jensen–Shannon divergence (JSD),
            JSD(p_test||p_ref)

            - 'mmd', the Maximum Mean Discrepancy (MMD),
            MMD(p_test, p_ref)

            - 'fd', the Frechet Distance (FD),
            FD(p_test, p_ref)

            - Callable function,
            Example: metric = roerich.scores.frechet_distance

        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for classification. In the most cases periods=1
        will be a good choice.

        window_size: int, default=100
            Number of consecutive observations of a time series in test and reference
        windows. Recommendation: select the value so that there is only one change point within 2*window_size
        observations of the signal.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_runs: int, default=1
            Number of times, the binary classifier runs on each pair of test and reference
        windows. Observations in the windows are divided randomly between train and validation sample every time.
        n_runs > 1 helps to reduce noise in the change point detection score.

        """

        super().__init__(periods, window_size, step, n_runs)

        nsam = min(10, window_size // 4 + 1)
        if base_classifier == 'qda':
            self.base_classifier = QuadraticDiscriminantAnalysis(store_covariance=True, reg_param=0.01)
        elif base_classifier == 'logreg':
            self.base_classifier = LogisticRegression()
        elif base_classifier == 'dt':
            self.base_classifier = DecisionTreeClassifier(min_samples_leaf=nsam, max_depth=6)
        elif base_classifier == 'rf':
            self.base_classifier = RandomForestClassifier(n_estimators=100, min_samples_leaf=nsam)
        elif base_classifier == 'mlp':
            self.base_classifier = MLPClassifier(hidden_layer_sizes=(100, 100), solver="adam", activation="relu",
                                                 learning_rate_init=0.1, max_iter=50, alpha=1.)
        elif base_classifier == 'knn':
            self.base_classifier = KNeighborsClassifier(n_neighbors=nsam)
        elif base_classifier == 'nb':
            self.base_classifier = GaussianNB(var_smoothing=0.01)
        else:
            self.base_classifier = base_classifier

        if metric == "KL" or metric == "kl":
            self.metric = KL
        elif metric == "KL_sym" or metric == "klsym":
            self.metric = KL_sym
        elif metric == "JSD" or metric == "jsd":
            self.metric = JSD
        elif metric == "PE" or metric == "pe":
            self.metric = PE
        elif metric == "PE_sym" or metric == "pesym":
            self.metric = PE_sym
        elif metric == "mmd":
            self.metric = maximum_mean_discrepancy
        elif metric == "fd":
            self.metric = frechet_distance
        else:
            self.metric = metric

    @ignore_warnings(category=ConvergenceWarning)
    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        X = np.vstack((X_ref, X_test))
        y = np.hstack((np.zeros(len(X_ref)), np.ones(len(X_test))))

        ss = StandardScaler()
        X = ss.fit_transform(X)

        X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                            test_size=0.5,
                                                            stratify=y,
                                                            random_state=np.random.randint(0, 1000))

        classifier = deepcopy(self.base_classifier)
        classifier.fit(X_train, y_train)
        y_pred = classifier.predict_proba(X_test)[:, 1]
        score = self.metric(y_pred[y_test == 0], y_pred[y_test == 1])

        return score


# NN RuLSIF
class ChangePointDetectionRuLSIF(ChangePointDetectionBase):

    def __init__(self, base_regressor='mlp', metric="pesym", periods=1, window_size=100, step=1, n_runs=1):
        """

        Change point detection algorithm based on RuLSIF regressor [1]. It takes two sliding windows
        (reference and test) in a signal, and directly estimates probability density ratio. The ratios are used to 
        calculate the change point detection score.

        [1] Mikhail Hushchyn and Andrey Ustyuzhanin. “Generalization of Change-Point Detection in Time Series Data Based on Direct Density Ratio Estimation.” J. Comput. Sci. 53 (2021): 101385.

        Parameters:
        -----------
        base_regressor: {'mlp', 'gbdt'} or callable, default='mlp'
            Sklearn-like regressor with the RuLSIF loss function to estimate the density ratio.

            - 'mlp', Multilayer Perceptron regressor with RuLSIF loss function,
              roerich.algorithms.models.NNRuLSIFRegressor(n_hidden=100, n_epochs=50, batch_size=2*window_size, lr=0.1, l2=0.01, alpha=0.05)
            
            - 'gbdt', Gradient Boosting over Decision Trees regressor with RuLSIF loss function,
              roerich.algorithms.models.GBDTRuLSIFRegressor(n_estimators=50, learning_rate=0.2, max_depth=3, alpha=0.05, min_samples_leaf=nsam)
              
            - Callable object,
            Example: base_regressor = roerich.algorithms.models.NNRuLSIFRegressor()

        metric: {'pesym'}, default='pesym'.
            Name of a score function, that is used to calculate the change point detection score based on predictions
            for reference (p_ref) and test (p_test) windows.

            - 'pesym' or 'PE_sym', Pearson (PE) divergence,
            PE(p_test||p_ref) + PE(p_ref||p_test)

        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for regression. In the most cases periods=1
        will be a good choice.

        window_size: int, default=100
            Number of consecutive observations of a time series in test and reference
        windows. Recommendation: select the value so that there is only one change point within 2*window_size
        observations of the signal.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_runs: int, default=1
            Number of times, the regressor runs on each pair of test and reference
        windows. Observations in the windows are divided randomly between train and validation sample every time.
        n_runs > 1 helps to reduce noise in the change point detection score.

        """

        super().__init__(periods, window_size, step, n_runs)

        nsam = min(10, window_size // 4 + 1)
        if base_regressor == 'gbdt':
            self.base_regressor = GBDTRuLSIFRegressor(n_estimators=50, learning_rate=0.2,
                                                      max_depth=3, alpha=0.05, min_samples_leaf=nsam)
        elif base_regressor == 'mlp':
            self.base_regressor = NNRuLSIFRegressor(n_hidden=100, n_epochs=50,
                                                    batch_size=2*window_size, lr=0.1, l2=0.01, alpha=0.05)
        else:
            self.base_regressor = base_regressor

        self.metric = metric

    def _one_side_predict(self, X_train, X_test, y_train, y_test):
        """
        Fit a regressor on a pair of the train sample and make a prediction on the test.

        Parameters:
        -----------
        X_train: numpy.ndarray
            Matrix of train observations.
        X_test: numpy.ndarray
            Matrix of test observations.
        y_train: numpy.array
            Array of train labels.
        y_test: numpy.array
            Array of test labels.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for the test.
        """

        reg = deepcopy(self.base_regressor)
        reg.fit(X_train, y_train)
        ratios = reg.predict_proba_ratio(X_test)

        if self.metric == "PE" or self.metric == "PE_sym" or self.metric == "pesym":
            score = 0.5 * np.mean(ratios[y_test == 1]) - 0.5
        else:
            score = 0

        return score

    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))

        ss = StandardScaler()
        X = ss.fit_transform(X)

        X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                            test_size=0.5,
                                                            stratify=y,
                                                            random_state=np.random.randint(0, 1000))

        score_right = self._one_side_predict(X_train, X_test, y_train, y_test)
        score_left = self._one_side_predict(X_test, X_train, 1 - y_test, 1 - y_train)
        score = score_right + score_left

        return score



import numpy as np
import pytest
from inspect import getmembers, isfunction
from roerich import scores

funcs = [f[1] for f in getmembers(scores, isfunction)]


@pytest.mark.parametrize("cost", funcs)
def test_same_in_size_2d(cost):
    N = 100
    X_real = np.random.multivariate_normal([0, 0], [[1, 0.7], [0.7, 1]], N)
    X_fake = np.random.multivariate_normal([0, 0.], [[1, 0.7], [0.7, 1]], N)
    score = cost(X_real, X_fake)

@pytest.mark.parametrize("cost", funcs)
def test_same_in_size_1d(cost):
    N = 100
    X_real = np.random.normal(0, 1, N).reshape(-1, 1)
    X_fake = np.random.normal(1, 1, N).reshape(-1, 1)
    score = cost(X_real, X_fake)

@pytest.mark.parametrize("cost", funcs)
def test_different_in_size_2d(cost):
    X_real = np.random.multivariate_normal([0, 0], [[1, 0.7], [0.7, 1]], 100)
    X_fake = np.random.multivariate_normal([0, 0.], [[1, 0.7], [0.7, 1]], 153)
    score = cost(X_real, X_fake)

    X_real = np.random.multivariate_normal([0, 0], [[1, 0.7], [0.7, 1]], 342)
    X_fake = np.random.multivariate_normal([0, 0.], [[1, 0.7], [0.7, 1]], 100)
    score = cost(X_real, X_fake)


import numpy as np
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler
from copy import deepcopy

from sklearn.utils._testing import ignore_warnings
from sklearn.exceptions import ConvergenceWarning

from .cpdc import ChangePointDetectionClassifier

# CV Classification
class ChangePointDetectionClassifierCV(ChangePointDetectionClassifier):
    
    def __init__(self, base_classifier='mlp', metric="klsym",
                 periods=1, window_size=100, step=1, n_runs=1, n_splits=5):
        """

        Change point detection algorithm based on binary classification [1] with Stratified KFold cross validation. 
        It takes to sliding windows (reference and test) in a signal, and separate them using the classifier. 
        The classification quality is considered as a change point detection score.

        [1] Mikhail Hushchyn and Andrey Ustyuzhanin. “Generalization of Change-Point Detection in Time Series Data Based on Direct Density Ratio Estimation.” J. Comput. Sci. 53 (2021): 101385.

        Parameters:
        -----------
        base_classifier: {'logreg', 'qda', 'dt', 'rf', 'mlp', 'knn', 'nb'} or callable, default='mlp'
            Sklearn-like binary classifier to separate reference and test sliding windows in the signal.

            - 'logreg', Logistic Regression classifier,
              sklearn.linear_model.LogisticRegression()

            - 'qda', Quadratic Discriminant Analysis classifier,
              sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis(store_covariance=True, reg_param=0.01)

            - 'dt', Decision Tree classifier,
              sklearn.tree.DecisionTreeClassifier(min_samples_leaf=10, max_depth=6)

            - 'rf', Random Forest classifier,
              sklearn.ensemble.RandomForestClassifier((n_estimators=100, min_samples_leaf=10))

            - 'mlp', Multilayer Perceptron classifier,
              sklearn.neural_network.MLPClassifier(hidden_layer_sizes=(100,100), solver="adam", activation="relu", learning_rate_init=0.1, max_iter=50, alpha=1.)

            - 'knn', K Neighbors classifier,
              sklearn.neighbors.KNeighborsClassifier(n_neighbors=10)

            - 'nb', Gaussian Naive Bayes classifier,
              sklearn.naive_bayes.GaussianNB(var_smoothing=0.01)

            - Callable sklearn-like classifier,
              Example: base_classifier = LogisticRegression()

        metric: {'klsym', 'pesym', 'jsd', 'mmd', 'fd'} or callable, default='klsym'.
            {'KL', 'PE'} will be deprecated in future versions.
            Name of a score function, that is used to measure the classifier quality based on predictions
            for reference (p_ref) and test (p_test) windows. It is considered as change point detection score.

            - 'klsym' or 'KL_sym', symmetric Kullback-Leibler (KL) divergence,
            KL(p_test||p_ref) + KL(p_ref||p_test)

            - 'pesym' or 'PE_sym', symmetric Pearson (PE) divergence,
            PE(p_test||p_ref) + PE(p_ref||p_test)

            - 'jsd' or 'JSD', Jensen–Shannon divergence (JSD),
            JSD(p_test||p_ref)

            - 'mmd', the Maximum Mean Discrepancy (MMD),
            MMD(p_test, p_ref)

            - 'fd', the Frechet Distance (FD),
            FD(p_test, p_ref)

            - Callable function,
            Example: metric = roerich.scores.frechet_distance

        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for classification. In the most cases periods=1
        will be a good choice.

        window_size: int, default=100
            Number of consecutive observations of a time series in test and reference
        windows. Recommendation: select the value so that there is only one change point within 2*window_size
        observations of the signal.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_runs: int, default=1
            Number of times, the binary classifier runs on each pair of test and reference
        windows. Observations in the windows are divided randomly between train and validation sample every time.
        n_runs > 1 helps to reduce noise in the change point detection score.
        
        n_splits: int, default=5
            Number of splits for KFold cross validation.

        """

        super().__init__(base_classifier=base_classifier, metric=metric, periods=periods,
                         window_size=window_size, step=step, n_runs=n_runs)
        self.n_splits = n_splits


    @ignore_warnings(category=ConvergenceWarning)
    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))

        kf = StratifiedKFold(n_splits=self.n_splits, shuffle=True, random_state=np.random.randint(0, 1000))
        y_pred_glob = []
        y_test_glob = []

        for i, (train_index, test_index) in enumerate(kf.split(X, y)):
            X_train, X_test, y_train, y_test = (X[train_index], 
                                                X[test_index], 
                                                y[train_index], 
                                                y[test_index])

            ss = StandardScaler()
            ss.fit(X_train)
            X_train = ss.transform(X_train)
            X_test = ss.transform(X_test)

            classifier = deepcopy(self.base_classifier)
            classifier.fit(X_train, y_train)
            y_pred = classifier.predict_proba(X_test)[:, 1]
            y_pred_glob.append(y_pred)
            y_test_glob.append(y_test)

        y_pred = np.concatenate(tuple(y_pred_glob), axis=0)
        y_test = np.concatenate(tuple(y_test_glob), axis=0)
        score = self.metric(y_pred[y_test == 0], y_pred[y_test == 1])

        return score


import numpy as np
from scipy.linalg import sqrtm


def frechet_distance(X, Y):
    '''
    Calculates the Frechet Distance between X and Y samples.

    Parameters:
    -----------
    X: numpy.ndarray of shape [n_samples, n_features]
        The first sample.
    Y: numpy.ndarray of shape [n_samples, n_features]
        The second sample.

    Return:
    -------
    distance: float
        The estimated Frechet Distance.
    '''

    # 1d array to nd one
    X = np.reshape(X, (len(X), -1))
    Y = np.reshape(Y, (len(Y), -1))

    X_mean, X_cov = X.mean(axis=0), np.cov(X, rowvar=False)
    Y_mean, Y_cov = Y.mean(axis=0), np.cov(Y, rowvar=False)

    if X.shape[1] == 1:
        X_cov = np.array([[X_cov]])
    if Y.shape[1] == 1:
        Y_cov = np.array([[Y_cov]])

    diff = np.sum((X_mean - Y_mean)**2.0)
    covmean, _ = sqrtm(X_cov.dot(Y_cov), disp=False)
    if np.iscomplexobj(covmean): covmean = covmean.real

    distance = diff + np.trace(X_cov) + np.trace(Y_cov) - 2 * np.trace(covmean)

    return distance



import numpy as np
import pytest
from inspect import getmembers, isfunction, isclass
from roerich import networks
import torch

models = [f[1] for f in getmembers(networks, isclass)]

@pytest.mark.parametrize("model", models)
def test_base(model):
    # generate samples
    n = 10
    n_inputs = 2
    X = torch.rand(size=(n, n_inputs))

    # make predictions
    reg = model(n_inputs=n_inputs)
    y_pred = reg(X)
    assert y_pred.size() == (n, 1)



from abc import ABCMeta, abstractmethod
import numpy as np
import pandas as pd
from sklearn.metrics import pairwise_distances, roc_curve, roc_auc_score
from sklearn.model_selection import StratifiedKFold, KFold, train_test_split
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from copy import deepcopy
from joblib import Parallel, delayed
from scipy import interpolate
from scipy.signal import argrelmax

from .cpdc import ChangePointDetectionBase, KL, KL_sym, JSD, PE, PE_sym

# EnergyDistance
class EnergyDistanceCalculator(ChangePointDetectionBase):


    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows by energy
        distance metric.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        n = X_ref.shape[0]
        E = 2*pairwise_distances(X_ref, X_test, metric='euclidean') - pairwise_distances(X_test, metric='euclidean') - pairwise_distances(X_ref, metric='euclidean')
        return np.sum(E)/n**2


from ..change_point.onnr import OnlineNNRuLSIF
from ..change_point.onnc import OnlineNNClassifier
from .cpdc import ChangePointDetectionClassifier, ChangePointDetectionRuLSIF
from .matrix import MatrixImportance
from .enrg_dist import EnergyDistanceCalculator
from .cpdc_cv import ChangePointDetectionClassifierCV
from .models import GBDTRuLSIFRegressor
from .models import RegressionNetwork, NNRuLSIFRegressor
from .models import ClassificationNetwork, NNClassifier


__all__ = [
    'OnlineNNClassifier',
    'OnlineNNRuLSIF',
    'ChangePointDetectionClassifier', 
    'ChangePointDetectionRuLSIF',
    'MatrixImportance',
    'EnergyDistanceCalculator',
    'ChangePointDetectionClassifierCV',
    'ClassificationNetwork', 
    'NNClassifier',
    'GBDTRuLSIFRegressor', 
    'RegressionNetwork', 
    'NNRuLSIFRegressor'
]






from .onnc import OnlineNNClassifier
from .onnr import OnlineNNRuLSIF
from ..algorithms.cpdc import ChangePointDetectionClassifier, ChangePointDetectionRuLSIF
from ..algorithms.cpdc_cv import ChangePointDetectionClassifierCV
from ..algorithms.enrg_dist import EnergyDistanceCalculator
from .windows import SlidingWindows


__all__ = [
    'OnlineNNClassifier', 
    'OnlineNNRuLSIF', 
    'ChangePointDetectionClassifier', 
    'ChangePointDetectionRuLSIF',
    'ChangePointDetectionClassifierCV',
    'EnergyDistanceCalculator',
    'SlidingWindows'
]






import numpy as np
import pandas as pd


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader
from sklearn.preprocessing import StandardScaler

device = 'cuda' if torch.cuda.is_available() else 'cpu'


class ClassificationNetwork(nn.Module):

    def __init__(self, n_inputs=1, n_hidden=10):
        """
        One-layer neural network for classification.

        Parameters:
        -----------
        n_inputs: int
            Number of inputs.
        n_hidden: int
            Number of neurons in a layer.
        """

        super(ClassificationNetwork, self).__init__()
        self.model = nn.Sequential(nn.Linear(n_inputs, n_hidden),
                                   nn.Tanh(),
                                   nn.Linear(n_hidden, 1),
                                   nn.Sigmoid()
                                   )

    def forward(self, x):
        return self.model(x)




class NNClassifier(object):

    def __init__(self, n_hidden=10, n_epochs=10, batch_size=64, lr=0.01, l2=0.):
        """
        Neural network classifier.

        Parameters:
        -----------
        n_hidden: int
            Number of neurons in the network.
        n_epochs: int
            Number of epochs to fit the classifier.
        batch_size: int
            Batch size used during the training of the classifier.
        lr: float
            Learning rate.
        l2: float
            Weight decay or L2 regularization parameter.
        """

        self.n_hidden = n_hidden
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.l2 = l2
        self.scaler = StandardScaler()
        self.class_ratio = 1


    def get_data_loader(self, X, y):

        X_tensor = torch.as_tensor(X, dtype=torch.float32, device=device)
        y_tensor = torch.as_tensor(y.reshape(-1, 1), dtype=torch.float32, device=device)
        data = TensorDataset(X_tensor, y_tensor)
        data_loader = DataLoader(data, batch_size=self.batch_size, shuffle=True)

        return data_loader


    def fit(self, X, y):
        """
        Fit the classifier.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).
        y: numpy.array
            Array of targets with shape (n_objects,).
        """

        y = np.array(y)
        self.class_ratio = (y == 1).sum() / (y == 0).sum()

        X_ss = self.scaler.fit_transform(X)
        train_loader = self.get_data_loader(X_ss, y)

        self.model = ClassificationNetwork(n_inputs=X_ss.shape[1], n_hidden=self.n_hidden)
        opt = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.l2)
        loss_func = nn.BCELoss()

        self.model.train(True)
        for epoch_i in range(self.n_epochs):
            for x_batch, y_batch in train_loader:
                y_pred_batch = self.model(x_batch)
                loss = loss_func(y_pred_batch, y_batch)
                opt.zero_grad()
                loss.backward()
                opt.step()


    def predict_proba(self, X):
        """
        Predicts class probabilities for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted class probabilities with shape (n_objects, 2).
        """

        X_ss = self.scaler.transform(X)
        X_tensor = torch.as_tensor(X_ss, dtype=torch.float32, device=device)

        self.model.train(False)
        y_pred = self.model(X_tensor)
        y_pred_1 = y_pred.cpu().detach().numpy()
        y_pred = np.hstack((1-y_pred_1, y_pred_1))

        return y_pred


    def predict_proba_ratio(self, X):
        """
        Predicts probability ratio for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted target values with shape (n_objects, 1).
        """

        eps = 10**-3
        proba_pos = self.predict_proba(X)[:, 1]
        ratios = (proba_pos / (1 - proba_pos + eps)) / self.class_ratio

        return ratios




# Regreesions
class RegressionNetwork(nn.Module):

    def __init__(self, n_inputs=1, n_hidden=10):
        """
        One-layer neural network for regression.

        Parameters:
        -----------
        n_inputs: int
            Number of inputs.
        n_hidden: int
            Number of neurons in a layer.
        """

        super(RegressionNetwork, self).__init__()
        self.model = nn.Sequential(nn.Linear(n_inputs, n_hidden),
                                   nn.Tanh(),
                                   nn.Linear(n_hidden, 1)
                                   )

    def forward(self, x):
        return self.model(x)





class NNRuLSIFRegressor(object):

    def __init__(self, n_hidden=10, n_epochs=10, batch_size=64, lr=0.01, l2=0., alpha=0.05):
        """
        Neural network regressor with RuLSIF loss function.

        Parameters:
        -----------
        n_hidden: int
            Number of neurons in the network.
        n_epochs: int
            Number of epochs to fit the regressor.
        batch_size: int
            Batch size used during the training of the regressor.
        lr: float
            Learning rate.
        l2: float
            Weight decay or L2 regularization parameter.
        alpha: float
            Parameter of the RuLSIF loss function.
        """

        self.n_hidden = n_hidden
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.l2 = l2
        self.alpha = alpha
        self.scaler = StandardScaler()


    def get_data_loader(self, X, y):

        X_tensor = torch.as_tensor(X, dtype=torch.float32, device=device)
        y_tensor = torch.as_tensor(y.reshape(-1, 1), dtype=torch.float32, device=device)
        data = TensorDataset(X_tensor, y_tensor)
        data_loader = DataLoader(data, batch_size=self.batch_size, shuffle=True)

        return data_loader


    def rulsif_loss(self, y_pred, y):

        y_pred_ref = y_pred[y == 0]
        y_pred_test = y_pred[y == 1]
        loss = 0.5 * (1 - self.alpha) * (y_pred_ref**2).mean() + \
               0.5 *      self.alpha  * (y_pred_test**2).mean() - (y_pred_test).mean()

        return loss


    def fit(self, X, y):
        """
        Fit the regressor.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).
        y: numpy.array
            Array of targets with shape (n_objects,).
        """

        X_ss = self.scaler.fit_transform(X)
        train_loader = self.get_data_loader(X_ss, y)

        self.model = RegressionNetwork(n_inputs=X_ss.shape[1], n_hidden=self.n_hidden)
        opt = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.l2)

        self.model.train(True)
        for epoch_i in range(self.n_epochs):
            for x_batch, y_batch in train_loader:
                y_pred_batch = self.model(x_batch)
                loss = self.rulsif_loss(y_pred_batch, y_batch)
                opt.zero_grad()
                loss.backward()
                opt.step()


    def predict(self, X):
        """
        Predicts target for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted target values with shape (n_objects, 1).
        """

        X_ss = self.scaler.transform(X)
        X_tensor = torch.as_tensor(X_ss, dtype=torch.float32, device=device)

        self.model.train(False)
        y_pred = self.model(X_tensor)
        y_pred = y_pred.cpu().detach().numpy()
        y_pred[y_pred <= 0] *= 0

        return y_pred


    def predict_proba_ratio(self, X):
        """
        Predicts probability ratio for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted target values with shape (n_objects, 1).
        """

        predictions = self.predict(X)

        return predictions




# GBDT-RuLSIF
from sklearn.tree import DecisionTreeRegressor
from sklearn.model_selection import train_test_split


class GBDTRuLSIFRegressor(object):

    def __init__(self, n_estimators=100, learning_rate=0.1, sample_frac=0.7, alpha=0.1,
                 max_depth=4, min_samples_leaf=1, min_samples_split=2, splitter='best',
                 max_features=None):

        """
        Gradient Boosting over Decision Trees regressor with RuLSIF loss function.

        Parameters:
        -----------
        n_estimators: int
            Number of trees.
        learning_rate: float
            Learning rate shrinks the contribution of each tree by `learning_rate`.
        sample_frac: float
            The fraction of samples to be used for fitting the individual tree.
        alpha: float
            Parameter of the RuLSIF loss function.
        max_depth: int
            Maximum depth of the individual regression tree.
        min_samples_leaf: int
            The minimum number of samples required to be at a leaf node of a tree.
        min_samples_split: int
            The minimum number of samples required to split an internal node of a tree.
        splitter: {“best”, “random”}, default=”best”
            Criterion used to split a node.
        max_features: {'auto', 'sqrt', 'log2'}, int or float
            The number of features to consider when looking for the best split.
        """

        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.sample_frac = sample_frac
        self.alpha = alpha
        self.max_depth = max_depth
        self.min_samples_leaf = min_samples_leaf
        self.min_samples_split = min_samples_split
        self.splitter = splitter
        self.max_features = max_features
        self.estimators = []


    def pe_loss(self, pred, y):

        L = (-0.5 *       self.alpha *  (pred)**2) * (y == 1) + \
            (-0.5 * (1. - self.alpha) * (pred)**2) * (y == 0) + pred * (y == 1) - 0.5

        return -L


    def pe_loss_grad(self, pred, y):

        dL = (-1. *       self.alpha  * (pred)) * (y == 1) + \
             (-1. * (1. - self.alpha) * (pred)) * (y == 0) + 1. * (y == 1)

        return -dL


    def fit(self, X, y):
        """
        Fit the regressor.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).
        y: numpy.array
            Array of targets with shape (n_objects,).
        """

        for i in range(self.n_estimators):
            X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=self.sample_frac)
            pred_train = self.predict(X_train)
            grad = - self.pe_loss_grad(pred_train, y_train)
            n1 = (y_train == 1).sum()
            n0 = (y_train == 0).sum()
            weights = ((y_train == 1) / n1 + (y_train == 0) / n0) * (n1 + n0)
            atree = DecisionTreeRegressor(max_depth=self.max_depth, min_samples_leaf=self.min_samples_leaf,
                                          min_samples_split=self.min_samples_split, splitter=self.splitter,
                                          max_features=self.max_features)
            atree.fit(X_train, grad, weights)
            self.estimators.append(atree)


    def predict(self, X):
        """
        Predicts target for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted target values with shape (n_objects, 1).
        """

        predictions = 1. + 0.1 * np.random.rand(len(X))
        for est in self.estimators:
            pred = est.predict(X)
            predictions += self.learning_rate * pred

        return predictions


    def predict_proba_ratio(self, X):
        """
        Predicts probability ratio for the objects.

        Parameters:
        -----------
        X: numpy.ndarray
            Matrix of input objects with shape (n_objects, n_features).

        Returns:
        --------
        y_pred: numpy.ndarray
            Predicted target values with shape (n_objects, 1).
        """

        predictions = self.predict(X)

        return predictions



import numpy as np
from .onnc import autoregression_matrix
from scipy.signal import argrelmax, savgol_filter
from scipy import interpolate

import torch
import torch.nn as nn
import torch.optim as optim
device = 'cuda' if torch.cuda.is_available() else 'cpu'


def SMA(scores, N):
    new_scores = []
    for i in range(0, len(scores)):
        s = i - N if i - N >= 0 else 0
        new_scores.append(np.mean(scores[s:i + 1], axis=0))
    return np.array(new_scores)


class BaseNN(nn.Module):
    def __init__(self, n_inputs=1):
        super(BaseNN, self).__init__()
        self.fc1 = nn.Linear(n_inputs, 100)
        self.act1 = nn.ReLU()
        self.fc2 = nn.Linear(100, 1)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act1(x)
        x = self.fc2(x)
        return x


class OnlineNNRuLSIF(object):

    def __init__(self, net='auto', scaler='auto', alpha=0.1, metric="pesym", window_size=1, lag_size=100, periods=1,
                 step=1, n_epochs=1, lr=0.1, lam=0., optimizer="RMSprop"):
        """

        Change point detection algorithm using RuLSIF regression based on online neural network [1].
        It takes two small (1-10) sliding windows (reference and test) in a signal, and directly estimates probability density ratio.
        The ratios are used to calculate the change point detection score.

        [1] Mikhail Hushchyn, Kenenbek Arzymatov and Denis Derkach. “Online Neural Networks for Change-Point Detection.” ArXiv abs/2010.01388 (2020)

        Parameters:
        -----------
        net: {'auto'} or callable, default='auto'
            PyTorch neural network for binary classification to separate reference and test sliding windows in the signal.

            - 'auto', two-layers neural network with 100 neurons in each layer, and ReLU activation.
              roerich.algorithms.onnr.BaseNN()

            - Callable, pytorch neural network for regression,
              Example: base_classifier = roerich.algorithms.onnr.BaseNN

        scaler: {'auto'}, default='auto'
            Ignored. It will be deprecated in future versions.

        metric: {'pesym'} or callable, default='pesym'.
            Name of a score function, that is used to measure the classifier quality based on predictions
            for reference (p_ref) and test (p_test) windows. It is considered as change point detection score.

            - 'pesym' or 'PE_sym', symmetric Pearson (PE) divergence,
            PE(p_test||p_ref) + PE(p_ref||p_test)

        window_size: int, default=1
            Number of consecutive observations of a time series in test and reference
        windows.

        lag_size: int, default=100
            Number of observations between the reference and test windows. Recommendation: select the value so that
        there is only one change point within 2*lag_size observations of the signal.

        periods: int, default=1
            Number of consecutive observations of a time series, considered as one input vector.
        The signal is considered as an autoregression process (AR) for classification. In the most cases periods=1
        will be a good choice.

        step: int, default=1
            Algorithm estimates change point detection score for each <step> observation. step > 1 helps
        to speed up the algorithm.

        n_epochs: int, default=1
            Number of training epochs per each pair of windows.

        lr: float, default=0.1
            Learning rate.

        lam: float, default=0
            L2 regularization.

        optimizer: {"Adam", "SGD", "RMSprop", "ASGD"}, default="RMSprop"
            Optimizer used to fit the online neural network.

        """

        if net == 'auto':
            self.base_net = BaseNN
        else:
            self.base_net = net

        self.net1 = None
        self.net2 = None
        self.alpha = alpha
        self.metric = metric
        self.window_size = window_size
        self.periods = periods
        self.lag_size = lag_size
        self.step = step
        self.n_epochs = n_epochs
        self.lr = lr
        self.lam = lam
        self.optimizer = optimizer

    def _init_networks(self, n_inputs):

        # init network
        self.net1 = self.base_net(n_inputs)
        self.net2 = self.base_net(n_inputs)

        # init optimizer
        if self.optimizer == "Adam":
            self.opt1 = torch.optim.Adam(self.net1.parameters(), lr=self.lr, weight_decay=self.lam)
            self.opt2 = torch.optim.Adam(self.net2.parameters(), lr=self.lr, weight_decay=self.lam)
        elif self.optimizer == "SGD":
            self.opt1 = torch.optim.SGD(self.net1.parameters(), lr=self.lr, weight_decay=self.lam)
            self.opt2 = torch.optim.SGD(self.net2.parameters(), lr=self.lr, weight_decay=self.lam)
        elif self.optimizer == "RMSprop":
            self.opt1 = torch.optim.RMSprop(self.net1.parameters(), lr=self.lr, weight_decay=self.lam)
            self.opt2 = torch.optim.RMSprop(self.net2.parameters(), lr=self.lr, weight_decay=self.lam)
        elif self.optimizer == "ASGD":
            self.opt1 = optim.ASGD(self.net1.parameters(), lr=self.lr, lambd=0.0, alpha=0.75, t0=0.0,
                                   weight_decay=self.lam)
            self.opt2 = optim.ASGD(self.net2.parameters(), lr=self.lr, lambd=0.0, alpha=0.75, t0=0.0,
                                   weight_decay=self.lam)
        else:
            self.opt1 = torch.optim.Adam(self.net1.parameters(), lr=self.lr, weight_decay=self.lam)
            self.opt2 = torch.optim.Adam(self.net2.parameters(), lr=self.lr, weight_decay=self.lam)

    def _rulsif_loss(self, y_pred_ref, y_pred_test, alpha):
        loss = 0.5 * (1 - alpha) * (y_pred_ref ** 2).mean() + \
               0.5 * alpha * (y_pred_test ** 2).mean() - (y_pred_test).mean()
        return loss

    def _postprocessing(self, T, T_score, score):
        """
        Interpolates and shifts a change point detection score, estimates peak positions.

        Parameters:
        -----------
        T: numpy.array
            A broader time-step interval
        T_score: numpy.array
            A time intervals of CPD scores
        score: numpy.array
            A CPD scores

        Returns:
        --------
        new_score: numpy.array
            Interpolated and shifted CPD scores.
        peaks: numpy.array
            Positions of peaks in the CPD scores.
        """

        shift = self.lag_size + self.window_size
        inter = interpolate.interp1d(T_score - shift, score,
                                     kind='linear', fill_value=(0, 0), bounds_error=False)
        score = inter(T)

        score = SMA(score, shift)

        length = max(min(21, shift // 2), 5)
        score = savgol_filter(score, length, 3)
        peaks = argrelmax(score, order=shift)[0]

        return score, peaks

    def predict(self, X, distance=5, height=None, smooth=False):
        """
        Estimate change point detection score for a time series.

        Parameters:
        -----------
        X: numpy.ndarray
            Time series observation.

        distance: int, default=5
            Ignored. It will be deprecated in future versions.

        height: float or None, default=None
            Ignored. It will be deprecated in future versions.

        smooth: {True, False}, default=False
            Ignored. It will be deprecated in future versions.

        Retunrs:
        --------
        T_score: numpy.array
            Array of timestamps.
        scores: numpy.array
            Estimated change point detection score.
        """

        # data preparation
        X_auto = autoregression_matrix(X, periods=self.periods, fill_value=0)
        T, reference, test = self._reference_test(X_auto)

        # init network
        self._init_networks(X_auto.shape[1])

        # cpd score
        score = [self._reference_test_predict(reference[i], test[i]) for i in range(len(reference))]
        T_score = np.array([T[i] for i in range(len(reference))])

        # cpd score postprocessing
        new_score, peaks = self._postprocessing(np.arange(len(X)), T_score, score)

        return new_score, peaks

    def _reference_test_predict(self, X_ref, X_test):

        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))

        X = torch.from_numpy(X).float()
        y = torch.from_numpy(y).float()

        n_last = min(self.window_size, self.step)
        self.net1.train(False)
        test_preds = self.net1(X[y == 1][-n_last:]).detach().numpy()

        self.net2.train(False)
        ref_preds = self.net2(X[y == 0][-n_last:]).detach().numpy()

        self.net1.train(True)
        self.net2.train(True)
        for epoch in range(self.n_epochs):  # loop over the dataset multiple times

            # forward + backward + optimize
            y_pred_batch = self.net1(X).squeeze()
            y_pred_batch_ref = y_pred_batch[y == 0]
            y_pred_batch_test = y_pred_batch[y == 1]
            loss = self._rulsif_loss(y_pred_batch_ref, y_pred_batch_test, self.alpha)
            self.opt1.zero_grad()
            loss.backward()
            self.opt1.step()

            # forward + backward + optimize
            y_pred_batch = self.net2(X).squeeze()
            y_pred_batch_ref = y_pred_batch[y == 1]
            y_pred_batch_test = y_pred_batch[y == 0]
            loss = self._rulsif_loss(y_pred_batch_ref, y_pred_batch_test, self.alpha)
            self.opt2.zero_grad()
            loss.backward()
            self.opt2.step()

        score = (0.5 * np.mean(test_preds) - 0.5) + (0.5 * np.mean(ref_preds) - 0.5)

        return score

    def _reference_test(self, X):
        N = self.lag_size
        ws = self.window_size
        T = []
        reference = []
        test = []
        for i in range(2 * ws + N - 1, len(X), self.step):
            T.append(i)
            reference.append(X[i - 2 * ws - N + 1:i - ws - N + 1])
            test.append(X[i - ws + 1:i + 1])
        return np.array(T), np.array(reference), np.array(test)



import numpy as np
import pytest
from roerich import metrics

labels_based = [metrics.false_positive_rate,
                metrics.true_positive_rate,
                metrics.precision_recall_scores,
                metrics.precision_score,
                metrics.recall_score]

@pytest.mark.parametrize("algo", labels_based)
def test_labels_based(algo):
    # generate samples
    cps_true = [100, 200, 300, 400, 500]
    cps_pred = [105, 230, 310, 350, 405, 490]

    # detection
    algo(cps_true, cps_pred)



score_based = [metrics.precision_recall_curve,
                metrics.pr_auc]

@pytest.mark.parametrize("algo", score_based)
def test_score_based(algo):
    # generate samples
    cps_true = [100, 200, 300, 400, 500]
    cps_pred = [105, 230, 310, 350, 405, 490]
    cps_score_pred = [1, 2, 3, 0.1, 5, 6]

    # detection
    algo(cps_true, cps_pred, cps_score_pred)



from ..algorithms.models import RegressionNetwork
from ..algorithms.models import ClassificationNetwork
from ..algorithms.net import MyNNRegressor, MyNN


__all__ = [
    'ClassificationNetwork',
    'RegressionNetwork',
    'MyNNRegressor',
    'MyNN'
]






import pandas as pd
import numpy as np
import torch
import roerich
from itertools import combinations


class MatrixImportance():

    def __init__(self, cpd):
        """
        Algorithm for calculating pairwise scores

        Parameters:
        -----------
        cpd: roerich.algorithms.ChangePointDetectionClassifier
            Change point detection algorithm 
        """

        self.cpd = cpd


    def predict(self, data):
        """
        Estimate change point detection score for a time series.
        
        Parameters:
        -----------
        data: numpy.array
            A broader time-step interval

        Retunrs:
        --------
        scores: numpy.array
            Estimated change point detection score.
        """

        return self.cpd.predict(data)


    def predict_union(self, data):
        """
        Calculates matrices of scores by pairwise adding features
        
        Parameters:
        -----------
        data: numpy.array
            A broader time-step interval

        Returns:
        --------
        matrix: numpy.array
            3d array - list of score matrices
        """

        if type(data) == pd.DataFrame: data = data.to_numpy()

        assert type(data) == np.ndarray, "'data' argument must be numpy array or pandas dataframe."
        assert data.shape[1] > 0, "'data' must contain at least 1 feature"

        single_FI = {}
        for feature in range(data.shape[1]):
            single_FI[feature], _ = self.cpd.predict(data[:, feature])

        pairs = list(combinations(np.arange(data.shape[1]), 2))

        paired_FI = {}
        for pair in pairs:
            paired_FI[pair], _ = self.cpd.predict(data[:, list(pair)])
            
            
        m, n = data.shape

        matrix = torch.empty(n, n, m)

        for feature_1 in range(n):
            for feature_2 in range(n): 

                if feature_1 != feature_2:
                    pair = None
                    if (feature_1, feature_2) in pairs: pair = (feature_1, feature_2)
                    if (feature_2, feature_1) in pairs: pair = (feature_2, feature_1)
                    c = paired_FI[pair]

                    a = single_FI[feature_1]
                    b = single_FI[feature_2]

                    matrix[feature_1, feature_2, :] = torch.tensor(c)
                else:
                    matrix[feature_1, feature_2, :] = torch.tensor(single_FI[feature_1])
        
        matrix = np.swapaxes(matrix.numpy(), 0, 2)
        
        return matrix


    def predict_exclude(self, data):
        """
        Calculates matrices of scores by pairwise deleting features
        
        Parameters:
        -----------
        data: numpy.array
            A broader time-step interval

        Returns:
        --------
        matrix: numpy.array
            3d array - list of score matricies 
        """

        if type(data) == pd.DataFrame: data = data.to_numpy()

        assert type(data) == np.ndarray, "'data' argument must be numpy array or pandas dataframe."
        assert data.shape[1] > 0, "'data' must contain at least 1 feature"

        single_FI = {}
        for feature in range(data.shape[1]):
            single_FI[feature], _ = self.cpd.predict(data[:, feature])

        pairs = list(combinations(np.arange(data.shape[1]), 2))

        paired_FI = {}
        for pair in pairs:
            paired_FI[pair], _ = self.cpd.predict(np.delete(data, list(pair), 1))
            
            
        m, n = data.shape

        matrix = torch.empty(n, n, m)

        for feature_1 in range(n):
            for feature_2 in range(n): 

                if feature_1 != feature_2:
                    pair = None
                    if (feature_1, feature_2) in pairs: pair = (feature_1, feature_2)
                    if (feature_2, feature_1) in pairs: pair = (feature_2, feature_1)
                    c = paired_FI[pair]

                    a = single_FI[feature_1]
                    b = single_FI[feature_2]

                    matrix[feature_1, feature_2, :] = torch.tensor(c)
                else:
                    matrix[feature_1, feature_2, :] = torch.tensor(single_FI[feature_1])
        
        matrix = np.swapaxes(matrix.numpy(), 0, 2)

        return matrix


